// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawnlibs.hpp"
#include "util/mathutils.hpp"
#include "util/flag.hpp"
#include "display/Texture.hpp"
#include "util/UsageLock.hpp"

#define NEW_TRUETYPE_CHAR_BEGIN 32
#define NEW_TRUETYPE_CHAR_END 192

#define NEW_TRUETYPE_VARIANT_BOLD FLAG_DEFINE(0)
#define NEW_TRUETYPE_VARIANT_ITALICS FLAG_DEFINE(1)

namespace Dawn {
  class NewTrueTypeFace;

  typedef uint32_t newtruetypelock_t;

  struct NewTrueTypeCharacter {
    int32_t advanceX;
    int32_t advanceY;
    glm::vec2 bitmapSize;
    glm::vec2 bitmapPosition;
    float_t textureY;
  };


  struct NewTrueTypeFaceTextureStyle {
    uint32_t fontSize;
    flag_t style;
  };

  class NewTrueTypeFaceTexture {
    protected:
      FT_Face *face;
      NewTrueTypeFace *trueTypeFace;
      std::map<FT_ULong, struct NewTrueTypeCharacter> characterData;
      struct NewTrueTypeFaceTextureStyle style;
      UsageLock locks;

    public:
      Texture texture;

      /**
       * Construct a new New True Type Face Texture object
       * 
       * @param trueTypeFace Face that this texture belongs to.
       * @param face The freetype face object.
       * @param style Style that this font has, used for locking.
       */
      NewTrueTypeFaceTexture(
        NewTrueTypeFace *trueTypeFace,
        FT_Face &face,
        struct NewTrueTypeFaceTextureStyle style
      );

      /**
       * Returns the character data for the given character.
       * 
       * @param c Character to get data for.
       * @return The Character data for the given character.
       */
      struct NewTrueTypeCharacter getCharacterData(FT_ULong c);

      /**
       * Destroys this true type face texture.
       */
      ~NewTrueTypeFaceTexture();

      friend class NewTrueTypeFace;
  };




  class NewTrueTypeFace {
    protected:
      std::vector<NewTrueTypeFaceTexture*> textures;
      FT_Face *face;
      std::map<struct NewTrueTypeFaceTextureStyle, NewTrueTypeFaceTexture*> texturesByStyle;
      std::map<struct NewTrueTypeFaceTextureStyle, std::vector<usagelockid_t>> locksByStyle;

    public:
      NewTrueTypeFace(FT_Face *face);
      usagelockid_t lock(struct NewTrueTypeFaceTextureStyle style);
      void unlock(usagelockid_t lock);
      NewTrueTypeFaceTexture * getTexture(usagelockid_t lock);
      ~NewTrueTypeFace();
  };
}